# -*- coding: utf-8 -*-
#
# dms_variants documentation build configuration file.
#
# This file is execfile()d with the current directory set to its
# containing dir.
#
# Note that not all possible configuration values are present in this
# autogenerated file.
#
# All configuration values have a default; values that are commented out
# serve to show the default.

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
"""Configuration file for sphinx."""

import datetime
import os
import re
import sys

sys.path.insert(0, '{0}/..'.format(os.path.abspath('.')))


# -- General configuration ------------------------------------------------

# If your documentation needs a minimal Sphinx version, state it here.
#
needs_sphinx = '1.0'

numfig = False

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
        'sphinx.ext.autodoc',
        'sphinx.ext.doctest',
        'sphinx.ext.mathjax',
        'sphinx.ext.githubpages',
        'sphinx.ext.viewcode',
        'sphinx.ext.napoleon',
        'matplotlib.sphinxext.plot_directive',
        'nbsphinx',
        'nbsphinx_link',
        ]

# configuration to matplotlib.sphinxext.plot_directive
plot_include_source = True
plot_html_show_source_link = False
# https://github.com/matplotlib/matplotlib/issues/4563#issuecomment-381366101
plot_rcparams = {'savefig.bbox': 'tight'}
plot_apply_rcparams = True

# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates']

# The suffix(es) of source filenames.
# You can specify multiple suffix as a list of string:
#
# source_suffix = ['.rst', '.md']
source_suffix = '.rst'

# The master toctree document.
master_doc = 'index'

# General information about the project.
project = u'dms_variants'
copyright = f"2019--{datetime.datetime.now().year}"  # noqa: A001
author = u'`the Bloom lab <https://research.fhcrc.org/bloom/en.html>`_'

# The version info for the project you're documenting, acts as replacement for
# |version| and |release|, also used in various other places throughout the
# built documents.
#
# The short X.Y version.
# Read in the version from file
versionfile = os.path.abspath(os.path.join(
        os.path.dirname(__file__), '../dms_variants/__init__.py'))
with open(versionfile) as f:
    versiontext = f.read()
versionmatches = re.findall(
            r'__version__\s+=\s+[\'"]([^\'"]+)[\'"]',
            versiontext)
if len(versionmatches) != 1:
    raise ValueError(f"found {len(versionmatches)} matches for version "
                     f"in {versionfile}")
else:
    versionstring = versionmatches[0]
# The full version, including alpha/beta/rc tags.
release = versionstring

# The language for content autogenerated by Sphinx. Refer to documentation
# for a list of supported languages.
#
# This is also used if you do content translation via gettext catalogs.
# Usually you set "language" from the command line for these cases.
language = None

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This patterns also effect to html_static_path and html_extra_path
exclude_patterns = ['_build', 'Thumbs.db', '.DS_Store', 'README.rst',
                    'doc_requirements.txt', '**.ipynb_checkpoints']

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = 'sphinx'

# If true, `todo` and `todoList` produce output, else they produce nothing.
todo_include_todos = False


# -- Options for HTML output ----------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = 'alabaster'

# Theme options are theme-specific and customize the look and feel of a theme
# further.  For a list of options available for each theme, see the
# documentation.
#
html_theme_options = {
        'logo': 'BloomLogo.jpg',
        'logo_name': 'true',
        'description': 'Analyze deep mutational scanning of barcoded variants',
        'github_button': 'true',
        'github_user': 'jbloomlab',
        'github_repo': 'dms_variants',
        'github_banner': 'true',
        'travis_button': 'true',
        'sidebar_width': '250px',
        }

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ['_static']

# Custom sidebar templates, must be a dictionary that maps document names
# to template names.
#
# This is required for the alabaster theme
# refs: http://alabaster.readthedocs.io/en/latest/installation.html#sidebars
html_sidebars = {
    '**': [
        'about.html',
        'navigation.html',
        'relations.html',  # needs 'show_related': True theme option to display
        'searchbox.html',
        'donate.html',
    ]
}


# -- Options for HTMLHelp output ------------------------------------------

# Output file base name for HTML help builder.
htmlhelp_basename = 'dms_variantsdoc'


# -- Options for LaTeX output ---------------------------------------------

latex_elements = {
    # The paper size ('letterpaper' or 'a4paper').
    #
    # 'papersize': 'letterpaper',

    # The font size ('10pt', '11pt' or '12pt').
    #
    # 'pointsize': '10pt',

    # Additional stuff for the LaTeX preamble.
    #
    # 'preamble': '',

    # Latex figure (float) alignment
    #
    # 'figure_align': 'htbp',
}

# Grouping the document tree into LaTeX files. List of tuples
# (source start file, target name, title,
#  author, documentclass [howto, manual, or own class]).
latex_documents = [
    (master_doc, 'dms_variants.tex', u'dms_variants Documentation',
     u'`the Bloom lab <https://research.fhcrc.org/bloom/en.html>`_', 'manual'),
]


# -- Options for manual page output ---------------------------------------

# One entry per manual page. List of tuples
# (source start file, name, description, authors, manual section).
man_pages = [
    (master_doc, 'dms_variants', u'dms_variants Documentation',
     [author], 1)
]


# -- Options for Texinfo output -------------------------------------------

# Grouping the document tree into Texinfo files. List of tuples
# (source start file, target name, title, author,
#  dir menu entry, description, category)
texinfo_documents = [
    (master_doc, 'dms_variants', u'dms_variants Documentation',
     author, 'dms_variants',
     'Analyze deep mutational scanning of barcoded variants',
     'Miscellaneous'),
    ]



"""
=================
binarymap
=================

Defines :class:`BinaryMap` objects for handling binary representations
of variants and their functional scores.

"""


import collections
import re

import numpy

import pandas as pd  # noqa: F401

import scipy.sparse

import dms_variants.constants


class BinaryMap:
    r"""Binary representations of variants and their functional scores.

    Note
    ----
    These maps represent variants as arrays of 0 and 1 integers indicating
    whether a particular variant has a substitution. The wildtype is all 0.
    Such representations are useful for fitting estimates of the effect of
    each substitution.

    Unless you are using the `expand` option, the binary maps only cover
    substitutions relative to wildtype that are present in at least one
    of the variants used to create the map.

    Parameters
    ----------
    func_scores_df : pandas.DataFrame
        Data frame of variants and their functional scores. Each row is
        a different variant, defined by space-delimited list of substitutions.
        Data frames of this type are returned by
        :meth:`dms_variants.codonvarianttable.CodonVariantTable.func_scores`.
    substitutions_col : str
        Column in `func_scores_df` giving substitutions for each variant.
    func_score_col : str or None
        Column in `func_scores_df` giving functional score for each variant,
        or `None` if no functional scores available.
    func_score_var_col : str or None
        Column in `func_scores_df` giving variance on functional score
        estimate, or `None` if no variance available.
    n_pre_col : str or None
        Column in `func_scores_df` giving pre-selection counts for each
        variant, or `None` if counts not available.
    n_post_col : str or None
        Column in `func_scores_df` giving post-selection counts for each
        variant, or `None` if counts not available.
    cols_optional : True
        All of the `*_col` parameters are optional except `substitutions_col`.
        If `cols_optional` is `True`, the absence of any of these columns
        is taken the same as setting that column's parameter to zero: the
        corresponding attribute is set to `None`.
    alphabet : list or tuple
        Allowed characters (e.g., amino acids or codons).
    expand : bool
        If `False` (the default) the encoding only covers substitutions
        relative to wildtype that are observed in the set of variants. If
        `True` then the encoding covers all allowed characters at each
        site regardless of whether they are wildtype or observed. In this
        latter case, each binary representation is of length (alphabet size)
        :math:`\times` (sequence length), and sums to the sequence length.
    wtseq : None or str
        Only set this option if `expand` is `True`. In that case, it
        should be the wildtype sequence.

    Attributes
    ----------
    binarylength : int
        Length of the binary representation of each variant.
    nvariants : int
        Number of variants.
    binary_variants : scipy.sparse.csr.csr_matrix of dtype int8
        Sparse matrix of shape `nvariants` by `binarylength`. Row
        `binary_variants[ivariant]` gives the binary representation of
        variant `ivariant`, and `binary_variants[ivariant, i]` is 1
        if the variant has the substitution :meth:`BinaryMap.i_to_sub`
        and 0 otherwise. To convert to dense `numpy.ndarray`, use
        `toarray` method of the sparse matrix.
    substitution_variants : list
        All variants as substitution strings as provided in `substitutions_col`
        of `func_scores_df`.
    func_scores : numpy.ndarray of floats
        A 1D array of length `nvariants` giving score for each variant.
    func_scores_var : numpy.ndarray of floats, or None
        A 1D array of length `nvariants` giving variance on score for each
        variant, or `None` if no variance estimates provided.
    n_pre : numpy.dnarray of integers, or None
        A 1D array of length `nvariants` giving pre-selection counts for each
        variant, or `None` if counts not provided.
    n_post : numpy.dnarray of integers, or None
        A 1D array of length `nvariants` giving post-selection counts for each
        variant, or `None` if counts not provided.
    alphabet : tuple
        Allowed characters (e.g., amino acids or codons).
    substitutions_col : str
        Value set when initializing object.

    Example
    -------
    Create a binary map:

    >>> func_scores_df = pd.DataFrame.from_records(
    ...         [('', 0.0, 0.2),
    ...          ('M1A', -0.2, 0.1),
    ...          ('M1C K3A', -0.4, 0.3),
    ...          ('', 0.01, 0.15),
    ...          ('A2C K3A', -0.05, 0.1),
    ...          ('A2*', -1.2, 0.4),
    ...          ],
    ...         columns=['aa_substitutions', 'func_score', 'func_score_var'])
    >>> binmap = BinaryMap(func_scores_df)

    The length of the binary representation equals the number of unique
    substitutions:

    >>> binmap.binarylength
    5
    >>> binmap.all_subs
    ['M1A', 'M1C', 'A2*', 'A2C', 'K3A']

    Scores, score variances, binary and string representations:

    >>> binmap.nvariants
    6
    >>> binmap.func_scores
    array([ 0.  , -0.2 , -0.4 ,  0.01, -0.05, -1.2 ])
    >>> binmap.func_scores_var
    array([0.2 , 0.1 , 0.3 , 0.15, 0.1 , 0.4 ])
    >>> type(binmap.binary_variants)
    <class 'scipy.sparse.csr.csr_matrix'>
    >>> binmap.binary_variants.toarray()
    array([[0, 0, 0, 0, 0],
           [1, 0, 0, 0, 0],
           [0, 1, 0, 0, 1],
           [0, 0, 0, 0, 0],
           [0, 0, 0, 1, 1],
           [0, 0, 1, 0, 0]], dtype=int8)
    >>> binmap.substitution_variants
    ['', 'M1A', 'M1C K3A', '', 'A2C K3A', 'A2*']
    >>> binmap.substitutions_col
    'aa_substitutions'

    Validate binary map interconverts binary representations and substitutions:

    >>> for ivar in range(binmap.nvariants):
    ...     binvar = binmap.binary_variants.toarray()[ivar]
    ...     subs_from_df = func_scores_df.at[ivar, 'aa_substitutions']
    ...     assert subs_from_df == binmap.binary_to_sub_str(binvar)
    ...     assert all(binvar == binmap.sub_str_to_binary(subs_from_df))

    Demonstrate :meth:`BinaryMap.sub_str_to_indices`:

    >>> for sub in binmap.substitution_variants:
    ...     print(binmap.sub_str_to_indices(sub))
    []
    [0]
    [1, 4]
    []
    [3, 4]
    [2]

    Now do similar operation but using `expand` to include full alphabet
    (although to keep size manageable, we use an alphabet smaller than
    all amino acids):

    >>> wtseq = 'MAKG'
    >>> alphabet = ['A', 'C', 'G', 'K', 'M', '*']
    >>> binmap_expand = BinaryMap(func_scores_df,
    ...                           alphabet=alphabet,
    ...                           expand=True,
    ...                           wtseq=wtseq)
    >>> binmap_expand.binarylength == len(wtseq) * len(alphabet)
    True

    >>> binmap_expand.all_subs
    ... # doctest: +NORMALIZE_WHITESPACE
    ['M1A', 'M1C', 'M1G', 'M1K', 'M1*',
     'A2C', 'A2G', 'A2K', 'A2M', 'A2*',
     'K3A', 'K3C', 'K3G', 'K3M', 'K3*',
     'G4A', 'G4C', 'G4K', 'G4M', 'G4*']

    >>> binmap_expand.binary_variants.toarray()
    ... # doctest: +NORMALIZE_WHITESPACE
    array([[0, 0, 0, 0, 1, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 1, 0,
            0, 0],
           [1, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 1, 0,
            0, 0],
           [0, 1, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 1, 0,
            0, 0],
           [0, 0, 0, 0, 1, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 1, 0,
            0, 0],
           [0, 0, 0, 0, 1, 0, 0, 1, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 1, 0,
            0, 0],
           [0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0, 0, 1, 0,
            0, 0]], dtype=int8)

    >>> all(numpy.sum(binmap_expand.binary_variants.toarray(), axis=1) ==
    ...     numpy.full(binmap_expand.nvariants, len(wtseq)))
    True

    >>> binmap_expand.substitution_variants
    ['', 'M1A', 'M1C K3A', '', 'A2C K3A', 'A2*']

    >>> for ivar in range(binmap_expand.nvariants):
    ...     binvar = binmap_expand.binary_variants.toarray()[ivar]
    ...     subs_from_df = func_scores_df.at[ivar, 'aa_substitutions']
    ...     assert subs_from_df == binmap_expand.binary_to_sub_str(binvar)
    ...     assert all(binvar == binmap_expand.sub_str_to_binary(subs_from_df))

    Note that `binamp` does not have `n_pre` and `n_post` attributes set:

    >>> binmap.n_pre == binmap.n_post == None
    True

    We would not have been able to initialize `binmap` if we weren't using
    the `cols_optional` flag:

    >>> BinaryMap(func_scores_df, alphabet=alphabet, cols_optional=False)
    Traceback (most recent call last):
      ...
    ValueError: `func_scores_df` lacks column pre_count

    Now assign values to `n_pre` and `n_post` attributes:

    >>> func_scores_df_counts = (
    ...         func_scores_df.assign(pre_count=[10, 20, 15, 5, 6, 8],
    ...                               post_count=[0, 3, 12, 11, 9, 8])
    ...         )
    >>> binmap_counts = BinaryMap(func_scores_df_counts, alphabet=alphabet)
    >>> binmap_counts.n_pre
    array([10, 20, 15,  5,  6,  8])
    >>> binmap_counts.n_post
    array([ 0,  3, 12, 11,  9,  8])

    """

    def __eq__(self, other):
        """Test if equal to object `other`.

        >>> df = pd.DataFrame({'aa_substitutions': ['', 'M1A'],
        ...                    'func_score': [0.0, -1.2],
        ...                    'func_score_var': [0.1, 0.15]})
        >>> df2 = df.copy()
        >>> df3 = df.assign(func_score=lambda x: x['func_score'] + 0.1)
        >>> bmap1 = BinaryMap(df)
        >>> bmap2 = BinaryMap(df2)
        >>> bmap3 = BinaryMap(df3)
        >>> bmap1 == bmap2
        True
        >>> bmap1 == bmap3
        False

        """
        # following here: https://stackoverflow.com/a/390640
        if type(other) != type(self):
            return False
        elif self.__dict__.keys() != other.__dict__.keys():
            return False
        else:
            for key, val in self.__dict__.items():
                val2 = getattr(other, key)
                if type(val) != type(val2):
                    return False
                elif isinstance(val, numpy.ndarray):
                    if not numpy.array_equal(val, val2):
                        return False
                elif isinstance(val, scipy.sparse.csr.csr_matrix):
                    if (val - val2).nnz:
                        return False
                elif isinstance(val, (pd.DataFrame, pd.Series)):
                    if not val.equals(val2):
                        return False
                else:
                    if val != val2:
                        return False
            return True

    def __init__(self,
                 func_scores_df,
                 *,
                 substitutions_col='aa_substitutions',
                 func_score_col='func_score',
                 func_score_var_col='func_score_var',
                 n_pre_col='pre_count',
                 n_post_col='post_count',
                 cols_optional=True,
                 alphabet=dms_variants.constants.AAS_WITHSTOP,
                 expand=False,
                 wtseq=None,
                 ):
        """Initialize object; see main class docstring."""
        self.nvariants = len(func_scores_df)
        self.alphabet = tuple(alphabet)

        for col, attr, dtype, lim_min, lim_max in [
                (func_score_col, 'func_scores', float, None, None),
                (func_score_var_col, 'func_scores_var', float, 0, None),
                (n_pre_col, 'n_pre', int, 0, None),
                (n_post_col, 'n_post', int, 0, None),
                ]:
            if col not in func_scores_df.columns:
                if cols_optional:
                    setattr(self, attr, None)
                else:
                    raise ValueError(f"`func_scores_df` lacks column {col}")
            else:
                vals = func_scores_df[col].values.astype(dtype)
                if not all(vals == func_scores_df[col].values):
                    raise ValueError(f"{col} not of type {dtype}")
                assert vals.shape == (self.nvariants,)
                if any(numpy.isnan(vals)):
                    raise ValueError(f"some entries in {col} are NaN")
                if (lim_min is not None) and any(vals < lim_min):
                    raise ValueError(f"some entries in {col} < {lim_min}")
                if (lim_max is not None) and any(vals > lim_max):
                    raise ValueError(f"some entries in {col} < {lim_min}")
                setattr(self, attr, vals)

        # get list of substitution strings for each variant
        if substitutions_col not in func_scores_df.columns:
            raise ValueError('`func_scores_df` lacks `substitutions_col` ' +
                             substitutions_col)
        substitutions = func_scores_df[substitutions_col].tolist()
        if not all(isinstance(s, str) for s in substitutions):
            raise ValueError('values in `substitutions_col` not all str')
        self.substitution_variants = substitutions
        self.substitutions_col = substitutions_col

        # regex that matches substitution
        chars = []
        for char in alphabet:
            if char.isalpha():
                chars.append(char)
            elif char == '*':
                chars.append(r'\*')
            else:
                raise ValueError(f"invalid alphabet character: {char}")
        chars = '|'.join(chars)
        self._sub_regex = (rf"(?P<wt>{chars})"
                           rf"(?P<site>\d+)"
                           rf"(?P<mut>{chars})")

        # build mapping from substitution to binary map index
        wts = {}
        muts = collections.defaultdict(set)
        for subs in substitutions:
            for sub in subs.split():
                m = re.fullmatch(self._sub_regex, sub)
                if not m:
                    raise ValueError(f"could not match substitution: {sub}")
                site = int(m.group('site'))
                if site not in wts:
                    wts[site] = m.group('wt')
                elif m.group('wt') != wts[site]:
                    raise ValueError(f"different wildtypes at {site}:\n"
                                     f"{m.group('wt')} versus {wts[site]}")
                if m.group('mut') == wts[site]:
                    raise ValueError(f"wildtype and mutant the same in {sub}")
                muts[site].add(m.group('mut'))
        self._i_to_sub = {}
        self._wt_indices = {}  # keyed by site, values wildtype indices
        if expand:
            if not isinstance(wtseq, str):
                raise ValueError('`wtseq` must be str if `expand` is True')
            if not set(wtseq).issubset(set(alphabet)):
                raise ValueError('`wtseq` has characters not in alphabet')
            if min(wts.keys()) < 1:
                raise ValueError('if `expand`, site numbers must start at 1')
            if max(wts.keys()) > len(wtseq):
                raise ValueError('`wtseq` not long enough given site numbers')
            for site, wt in wts.items():
                if wtseq[site - 1] != wt:
                    raise ValueError('`wtseq` and `func_scores_df` differ on '
                                     f"identity at site {site}")
            i = 0
            for site, wt in enumerate(wtseq, start=1):
                assert (site not in wts) or (wts[site] == wt)
                for char in self.alphabet:
                    self._i_to_sub[i] = f"{wt}{site}{char}"
                    if char == wt:
                        assert site not in self._wt_indices
                        self._wt_indices[site] = i
                    i += 1
        else:
            if wtseq is not None:
                raise ValueError('`wtseq` should be None if `expand` is False')
            i = 0
            for site, wt in sorted(wts.items()):
                for mut in sorted(muts[site]):
                    self._i_to_sub[i] = f"{wt}{site}{mut}"
                    i += 1
        self.binarylength = len(self._i_to_sub)
        self._sub_to_i = {sub: i for i, sub in self._i_to_sub.items()}
        self._wt_index_set = set(self._wt_indices.values())
        assert len(self._sub_to_i) == len(self._i_to_sub) == self.binarylength

        # build binary_variants
        row_ind = []  # row indices of elements that are one
        col_ind = []  # column indices of elements that are one
        for ivariant, subs in enumerate(substitutions):
            for isub in self.sub_str_to_indices(subs):
                row_ind.append(ivariant)
                col_ind.append(isub)
        self.binary_variants = scipy.sparse.csr_matrix(
                        (numpy.ones(len(row_ind), dtype='int8'),
                         (row_ind, col_ind)),
                        shape=(self.nvariants, self.binarylength),
                        dtype='int8')

    def sub_str_to_binary(self, sub_str):
        """Convert space-delimited substitutions to binary representation.

        Parameters
        ----------
        sub_str : str
            Space-delimited substitutions.

        Returns
        -------
        numpy.ndarray of dtype `int8`
            Binary representation.

        """
        binrep = numpy.zeros(self.binarylength, dtype='int8')
        binrep[self.sub_str_to_indices(sub_str)] = 1
        return binrep

    def sub_str_to_indices(self, sub_str):
        """Convert space-delimited substitutions to list of non-zero indices.

        Parameters
        -----------
        sub_str : str
            Space-delimited substitutions.

        Returns
        -------
        list
            Contains binary representation index for each mutation, so wildtype
            is an empty list.

        """
        sites = set()
        indices = []
        for sub in sub_str.split():
            m = re.fullmatch(self._sub_regex, sub)
            if not m:
                raise ValueError(f"substitution {sub} in {sub_str} invalid "
                                 f"for alphabet {self.alphabet}")
            if m.group('site') in sites:
                raise ValueError("multiple subs at same site in {sub_str}")
            sites.add(m.group('site'))
            indices.append(self.sub_to_i(sub))
        sites = {int(site) for site in sites}
        for site, i in self._wt_indices.items():
            if site not in sites:
                indices.append(i)
        return sorted(indices)

    def binary_to_sub_str(self, binary):
        """Convert binary representation to space-delimited substitutions.

        Note
        ----
        This method is the inverse of :meth:`BinaryMap.sub_str_to_binary`.

        Parameters
        ----------
        binary : numpy.ndarray
            Binary representation.

        Returns
        -------
        str
            Space-delimited substitutions.

        """
        if binary.shape != (self.binarylength,):
            raise ValueError(f"`binary` not length {self.binarylength}:\n" +
                             str(binary))
        if not set(binary).issubset({0, 1}):
            raise ValueError(f"`binary` not all 0 or 1:\n{binary}")
        subs = [s for s in map(self.i_to_sub, numpy.flatnonzero(binary)) if s]
        sites = [re.fullmatch(self._sub_regex, sub) for sub in subs]
        if len(sites) != len(set(sites)):
            raise ValueError('`binary` specifies multiple substitutions '
                             f"at same site:\n{binary}\n{' '.join(subs)}")
        return ' '.join(subs)

    def i_to_sub(self, i):
        """Substitution corresponding to index in binary representation.

        Parameters
        ----------
        i : int
            Index in binary representation, 0 <= `i` < `binarylength`.

        Returns
        -------
        str
            The substitution corresponding to that index.

        """
        try:
            if i in self._wt_index_set:
                return ''
            else:
                return self._i_to_sub[i]
        except KeyError:
            if i < 0 or i >= self.binarylength:
                raise ValueError(f"invalid i of {i}. Must be >= 0 and "
                                 f"< {self.binarylength}")
            else:
                raise ValueError(f"unexpected error, i = {i} should be in map")

    def sub_to_i(self, sub):
        """Index in binary representation corresponding to substitution.

        Parameters
        ----------
        sub : str
            The substitution.

        Returns
        -------
        int
            Index in binary representation, will be >= 0 and < `binarylength`.

        """
        try:
            return self._sub_to_i[sub]
        except KeyError:
            raise ValueError(f"sub of {sub} is not in the binary map. The map "
                             'only contains substitutions in the variants.')

    @property
    def all_subs(self):
        """list: Substitutions in order encoded in binary map."""
        if not hasattr(self, '_all_subs'):
            self._all_subs = [self.i_to_sub(i) for i in
                              range(self.binarylength) if self.i_to_sub(i)]
        return self._all_subs


if __name__ == '__main__':
    import doctest
    doctest.testmod()



"""
========
barcodes
========

Utility functions to process and analyze barcodes.

"""

import collections
import math
import random  # noqa: F401

import numpy

import pandas as pd

import scipy.special


def rarefyBarcodes(barcodecounts, *,
                   barcodecol='barcode', countcol='count',
                   maxpoints=100000, logspace=True):
    """Rarefaction curve of barcode observations.

    Note
    ----
    Uses analytical formula for rarefaction defined here:
    https://en.wikipedia.org/wiki/Rarefaction_(ecology)#Derivation

    Parameters
    ----------
    barcodecounts :pandas.DataFrame
        Data frame with counts to rarefy.
    barcodecol : str
        Column in `barcodecounts` listing all unique barcodes.
    countcol : str
        Column in `barcodecounts` with observed counts of each barcode.
    maxpoints : int
        Only calculate rarefaction curve at this many points. Benefit
        is that it is costly to calculate the curve for many points.
    logspace : bool
        Logarithmically space the points. If `False`, space linearly.

    Returns
    -------
    pandas.DataFrame
        A data frame with columns 'ncounts' and 'nbarcodes' giving number
        of unique barcodes observed for each total number of observed counts.

    Example
    -------
    >>> barcodecounts = pd.DataFrame({'barcode': ['A', 'G', 'C', 'T'],
    ...                               'count': [4, 2, 1, 0]})
    >>> rarefaction_curve = rarefyBarcodes(barcodecounts)
    >>> rarefaction_curve
       ncounts  nbarcodes
    0        1   1.000000
    1        2   1.666667
    2        3   2.114286
    3        4   2.428571
    4        5   2.666667
    5        6   2.857143
    6        7   3.000000

    Verify this result matches what is obtained by random sampling:

    >>> random.seed(1)
    >>> barcodelist = []
    >>> for tup in barcodecounts.itertuples(index=False):
    ...     barcodelist += [tup.barcode] * tup.count
    >>> nrand = 10000
    >>> ncounts = list(range(1, barcodecounts['count'].sum() + 1))
    >>> nbarcodes = []
    >>> for ncount in ncounts:
    ...     nbarcodes.append(sum(len(set(random.sample(barcodelist, ncount)))
    ...                      for _ in range(nrand)) / nrand)
    >>> sim_rarefaction_curve = pd.DataFrame({'ncounts': ncounts,
    ...                                       'nbarcodes': nbarcodes})
    >>> numpy.allclose(rarefaction_curve, sim_rarefaction_curve, atol=1e-2)
    True

    """
    if len(barcodecounts) != len(barcodecounts[barcodecol].unique()):
        raise ValueError('non-unique barcodes in `barcodecounts`')

    # follow nomenclature at
    # https://en.wikipedia.org/wiki/Rarefaction_(ecology)#Derivation
    Ni = barcodecounts.set_index(barcodecol)[countcol].to_dict()
    N = sum(Ni.values())
    K = len(barcodecounts)
    Mj = collections.Counter(Ni.values())

    Nk, num = map(numpy.array, zip(*Mj.items()))

    # use simplification that (N - Ni)Cr(n) / (N)Cr(n) =
    # [(N - Ni)! * (N - n)!] / [N! * (N - Ni - n)!]
    #
    # Also use fact that gamma(x + 1) = x!
    nbarcodes = []
    lnFactorial_N = scipy.special.gammaln(N + 1)
    if logspace and N > maxpoints:
        ncounts = list(numpy.unique(numpy.logspace(
                       math.log10(1), math.log10(N),
                       num=min(N, maxpoints)).astype('int')))
    else:
        ncounts = list(numpy.unique(numpy.linspace(
                       1, N, num=min(N, maxpoints)).astype('int')))
    for n in ncounts:
        lnFactorial_N_minus_n = scipy.special.gammaln(N - n + 1)
        i = numpy.nonzero(N - Nk - n >= 0)  # indices where this is true
        nbarcodes.append(
                K - (num[i] * numpy.exp(
                            scipy.special.gammaln(N - Nk[i] + 1) +
                            lnFactorial_N_minus_n -
                            lnFactorial_N -
                            scipy.special.gammaln(N - Nk[i] - n + 1))
                     ).sum()
                )
    return pd.DataFrame({'ncounts': ncounts, 'nbarcodes': nbarcodes})


if __name__ == '__main__':
    import doctest
    doctest.testmod()



"""
===========
pdb_utils
===========

Functions to manipulate `PDB <https://www.rcsb.org/>`_ files.

"""


import collections  # noqa: F401
import os  # noqa: F401
import tempfile  # noqa: F401
import warnings

import Bio.PDB

import pandas as pd  # noqa: F401

import requests  # noqa: F401


def reassign_b_factor(input_pdbfile,
                      output_pdbfile,
                      df,
                      metric_col,
                      *,
                      site_col='site',
                      chain_col='chain',
                      missing_metric=0,
                      model_index=0,
                      ):
    r"""Reassign B factors in PDB file to some other metric.

    B-factor re-assignment is useful because PDB images can be colored
    by B factor using programs such as ``pymol`` using commands like::

        show surface, RBD; spectrum b, white red, RBD, minimum=0, maximum=1

    Parameters
    ----------
    input_pdbfile : str
        Path to input PDB file.
    output_pdbfile: str
        Name of created output PDB file with re-assigned B factors.
    df : pandas.DataFrame
        Data frame with metric used to re-assign B factor.
    metric_col : str
        Name of column in `df` that has the numerical metric that the B
        factor is re-assigned to.
    site_col : str
        Name of column in `df` with site numbers, which should map numbers
        used in PDB.
    chain_col : str
        Name of column in `df` with chain labels.
    missing_metric : float or dict
        How do we handl sites that are missing in `df`? If a float, reassign
        B factors for all missing sites to this value. If a dict, should be
        keyed by chain and assign all missing sites in each chain to
        indicated value.
    model_index : int
        Which model in the PDB to use. If a X-ray structure, there is
        probably just one model so you can use default of 0.

    Returns
    -------
    None

    Example
    -------
    Create data frame `df` that assigns metric to two sites in chain E:

    >>> df = pd.DataFrame({'chain': ['E', 'E'],
    ...                    'site': [333, 334],
    ...                    'metric': [0.5, 1.2]})

    Create dict `missing_metric` that assigns -1 to sites with missing
    metrics in chain E, and 0 to sites in other chains:

    >>> missing_metric = collections.defaultdict(lambda: 0)
    >>> missing_metric['E'] = -1

    Download PDB, do the re-assignment of B factors, read the lines
    from the resulting re-assigned PDB:

    >>> pdb_url = 'https://files.rcsb.org/download/6M0J.pdb'
    >>> r = requests.get(pdb_url)
    >>> with tempfile.TemporaryDirectory() as tmpdir:
    ...    original_pdbfile = os.path.join(tmpdir, 'original.pdb')
    ...    with open(original_pdbfile, 'wb') as f:
    ...        _ = f.write(r.content)
    ...    reassigned_pdbfile = os.path.join(tmpdir, 'reassigned.pdb')
    ...    reassign_b_factor(input_pdbfile=original_pdbfile,
    ...                      output_pdbfile=reassigned_pdbfile,
    ...                      df=df,
    ...                      metric_col='metric',
    ...                      missing_metric=missing_metric)
    ...    pdb_text = open(reassigned_pdbfile).readlines()

    Now spot check some key lines in the output PDB.
    Chain A has all sites with B factors (last entry) re-assigned to 0:

    >>> print(pdb_text[0].strip())
    ATOM      1  N   SER A  19     -31.455  49.474   2.505  1.00  0.00           N

    Chain E has sites 333 and 334 with B-factors assigned to values in `df`, and
    other sites (such as 335) assigned to -1:

    >>> print('\n'.join(line.strip() for line in pdb_text[5010: 5025]))
    ATOM   5010  O   THR E 333     -34.954  13.568  46.370  1.00  0.50           O
    ATOM   5011  CB  THR E 333     -33.695  14.409  48.627  1.00  0.50           C
    ATOM   5012  OG1 THR E 333     -34.797  14.149  49.507  1.00  0.50           O
    ATOM   5013  CG2 THR E 333     -32.495  14.879  49.438  1.00  0.50           C
    ATOM   5014  N   ASN E 334     -35.532  15.604  45.605  1.00  1.20           N
    ATOM   5015  CA  ASN E 334     -36.287  15.087  44.474  1.00  1.20           C
    ATOM   5016  C   ASN E 334     -35.475  15.204  43.182  1.00  1.20           C
    ATOM   5017  O   ASN E 334     -34.533  15.994  43.076  1.00  1.20           O
    ATOM   5018  CB  ASN E 334     -37.622  15.823  44.337  1.00  1.20           C
    ATOM   5019  CG  ASN E 334     -38.660  15.006  43.586  1.00  1.20           C
    ATOM   5020  OD1 ASN E 334     -38.568  13.776  43.514  1.00  1.20           O
    ATOM   5021  ND2 ASN E 334     -39.649  15.686  43.016  1.00  1.20           N
    ATOM   5022  N   LEU E 335     -35.849  14.391  42.194  1.00 -1.00           N
    ATOM   5023  CA  LEU E 335     -35.084  14.305  40.955  1.00 -1.00           C
    ATOM   5024  C   LEU E 335     -35.466  15.426  39.992  1.00 -1.00           C

    """  # noqa: E501
    # subset `df` to needed columns and error check it
    cols = [metric_col, site_col, chain_col]
    for col in cols:
        if col not in df.columns:
            raise ValueError(f"`df` lacks column {col}")
    df = df[cols].drop_duplicates()
    if len(df) != len(df.groupby([site_col, chain_col])):
        raise ValueError('non-unique metric for a site in a chain')

    if df[site_col].dtype != int:
        raise ValueError('function currently requires `site_col` to be int')

    # read PDB, catch warnings about discontinuous chains
    with warnings.catch_warnings():
        warnings.simplefilter(
                'ignore',
                category=Bio.PDB.PDBExceptions.PDBConstructionWarning)
        pdb = Bio.PDB.PDBParser().get_structure('_', input_pdbfile)

    # get the model out of the PDB
    model = list(pdb.get_models())[model_index]

    # make sure all chains in PDB
    missing_chains = set(df[chain_col]) - {chain.id for chain
                                           in model.get_chains()}
    if missing_chains:
        raise ValueError(f"`df` has chains not in PDB: {missing_chains}")

    # make missing_metric a dict if it isn't already
    if not isinstance(missing_metric, dict):
        missing_metric = {chain.id: missing_metric for chain
                          in model.get_chains()}

    # loop over all chains and do coloring
    for chain in model.get_chains():
        chain_id = chain.id
        site_to_val = (df
                       .query(f"{chain_col} == @chain_id")
                       .set_index(site_col)
                       [metric_col]
                       .to_dict()
                       )
        for residue in chain:
            site = residue.get_id()[1]
            try:
                metric_val = site_to_val[site]
            except KeyError:
                metric_val = missing_metric[chain_id]
            # for disordered residues, get list of them
            try:
                residuelist = residue.disordered_get_list()
            except AttributeError:
                residuelist = [residue]
            for r in residuelist:
                for atom in r:
                    # for disordered atoms, get list of them
                    try:
                        atomlist = atom.disordered_get_list()
                    except AttributeError:
                        atomlist = [atom]
                    for a in atomlist:
                        a.bfactor = metric_val

    # write PDB
    io = Bio.PDB.PDBIO()
    io.set_structure(pdb)
    io.save(output_pdbfile)


if __name__ == '__main__':
    import doctest
    doctest.testmod()



"""Tests counting codon variants.

Runs on a snippet of real data for RecA.

This test doesn't test for correct output, but rather that output is the
same as a version of the code that we are confident is correct. Therefore,
if later changes break this code, this test will identify that fact.

Essentially, it tests `CodonVariantTable` and `IlluminBarcodeParser`.

Written by Jesse Bloom.
"""


import itertools
import os
import unittest

import Bio.SeqIO

import pandas as pd
from pandas.testing import assert_frame_equal

from dms_variants.codonvarianttable import CodonVariantTable
from dms_variants.illuminabarcodeparser import IlluminaBarcodeParser


class test_count_codonvariants(unittest.TestCase):
    """Test counting of codon variants."""

    def test_count_codonvariants(self):
        """Test counting of codon variants."""
        indir = os.path.join(os.path.abspath(os.path.dirname(__file__)),
                             'count_codonvariant_files/')
        outdir = os.path.join(os.path.abspath(os.path.dirname(__file__)),
                              'test_count_codonvariants_files/')
        os.makedirs(outdir, exist_ok=True)

        barcode_variant_file = os.path.join(indir, 'barcode_variant_table.csv')

        ampliconfile = os.path.join(indir, 'PacBio_amplicon.gb')
        amplicon = Bio.SeqIO.read(ampliconfile, 'genbank')

        gene = [f for f in amplicon.features if f.type == 'gene']
        assert len(gene) == 1, "Failed to find exactly one gene feature"
        geneseq = str(gene[0].location.extract(amplicon).seq)

        variants = CodonVariantTable(
                            barcode_variant_file=barcode_variant_file,
                            geneseq=geneseq)
        variants2 = CodonVariantTable(
                            barcode_variant_file=barcode_variant_file,
                            geneseq=geneseq)

        n_variants_file = os.path.join(indir, 'n_variants.csv')
        assert_frame_equal(
                variants.n_variants_df(samples=None)
                        .assign(library=lambda x: x['library'].astype('str'),
                                sample=lambda x: x['sample'].astype('str')
                                ),
                pd.read_csv(n_variants_file),
                check_dtype=False
                )

        for mut_type in ['aa', 'codon']:
            _ = variants.plotNumMutsHistogram(mut_type, samples=None,
                                              max_muts=7)

        for variant_type in ['all', 'single']:
            _ = variants.plotNumCodonMutsByType(variant_type, samples=None,)

        _ = variants.plotNumCodonMutsByType('all', samples=None, min_support=2)

        for variant_type, mut_type in itertools.product(['all', 'single'],
                                                        ['aa', 'codon']):
            _ = variants.plotCumulMutCoverage(variant_type,
                                              mut_type,
                                              samples=None)

        for variant_type, mut_type in itertools.product(['all', 'single'],
                                                        ['aa', 'codon']):
            _ = variants.plotMutFreqs(variant_type, mut_type, samples=None)

        for mut_type in ['aa', 'codon']:
            _ = variants.plotMutHeatmap('all', mut_type,
                                        samples=None, libraries='all_only',
                                        widthscale=2)

        samples = (pd.DataFrame({
                        'library': ['library-1', 'library-2', 'library-1'],
                        'sample': ['plasmid', 'plasmid', 'uninduced'],
                        })
                   .assign(
                        run='run-1',
                        upstream='TTTTTAAGTTGTAAGGATATGCCATTCTAGA',
                        downstream='',
                        R1file=lambda x: (indir + x['library'] + '_' +
                                          x['sample'] + '_R1.fastq'),
                        R2file=lambda x: (indir + x['library'] + '_' +
                                          x['sample'] + '_R2.fastq'),
                        )
                   )

        barcode = [f for f in amplicon.features if f.type == 'barcode']
        assert len(barcode) == 1, "Failed to find exactly one barcode feature"
        bclen = len(barcode[0])

        fates = []
        counts_df = []
        for (lib, sample), runs in samples.groupby(['library', 'sample']):

            # read barcodes for all runs for library / sample
            barcodes = []
            for run_tup in runs.itertuples(index=False):
                parser = IlluminaBarcodeParser(
                            bclen=bclen,
                            upstream=run_tup.upstream,
                            downstream=run_tup.downstream,
                            valid_barcodes=variants.valid_barcodes(lib),
                            upstream_mismatch=3,
                            downstream_mismatch=3)
                run_barcodes, run_fates = parser.parse(
                                    r1files=run_tup.R1file,
                                    r2files=None)
                barcodes.append(run_barcodes)
                fates.append(run_fates.assign(library=lib,
                                              sample=sample,
                                              run=run_tup.run
                                              ))

            # combine barcodes read for each run
            barcodes = (pd.concat(barcodes, ignore_index=True, sort=False)
                        .groupby('barcode')
                        .aggregate({'count': 'sum'})
                        .reset_index()
                        )

            counts_df.append(barcodes.assign(library=lib, sample=sample))

            # add barcode counts to variant table
            variants.addSampleCounts(lib, sample, barcodes)

        # make sure adding by `add_sample_counts_df` is same
        counts_df = pd.concat(counts_df)
        variants2.add_sample_counts_df(counts_df)
        assert_frame_equal(variants.variant_count_df,
                           variants2.variant_count_df,
                           check_like=True)

        # concatenate read fates into one data frame
        fates = (pd.concat(fates, ignore_index=True, sort=False)
                 .assign(count=lambda x: x['count'].fillna(0).astype('int'))
                 )

        fatesfile = os.path.join(indir, 'fates.csv')
        assert_frame_equal(
                fates,
                pd.read_csv(fatesfile)
                )

        libs_to_analyze = ['library-1']

        for mut_type in ['aa', 'codon']:
            _ = variants.plotNumMutsHistogram(mut_type,
                                              libraries=libs_to_analyze,
                                              max_muts=7, orientation='v')

        for variant_type in ['all', 'single']:
            _ = variants.plotNumCodonMutsByType(variant_type,
                                                libraries=libs_to_analyze,
                                                orientation='v')

        for variant_type, mut_type in itertools.product(['all', 'single'],
                                                        ['aa', 'codon']):
            _ = variants.plotCumulMutCoverage(variant_type, mut_type,
                                              libraries=libs_to_analyze,
                                              orientation='v')

        for variant_type, mut_type in itertools.product(['all', 'single'],
                                                        ['aa', 'codon']):
            _ = variants.plotMutFreqs(variant_type, mut_type,
                                      libraries=libs_to_analyze,
                                      orientation='v')

        for mut_type in ['aa', 'codon']:
            _ = variants.plotMutHeatmap('all', mut_type,
                                        libraries=libs_to_analyze,
                                        widthscale=2)

        variant_count_file = os.path.join(indir, 'variant_counts.csv')
        assert_frame_equal(
            variants.variant_count_df
                    .assign(library=lambda x: x['library'].astype(str),
                            sample=lambda x: x['sample'].astype(str)
                            ),
            pd.read_csv(variant_count_file).fillna('')
            )

        countfiles = []
        for mut_type in ['single', 'all']:
            countsdir = os.path.join(outdir, f'{mut_type}_mutant_codoncounts')
            mut_type_countfiles = variants.writeCodonCounts(mut_type,
                                                            outdir=countsdir)
            countfiles.append(mut_type_countfiles
                              .assign(mutant_type=f'{mut_type} mutants')
                              )


if __name__ == '__main__':
    runner = unittest.TextTestRunner()
    unittest.main(testRunner=runner)
